/*
 * shrip - A DVD Encoder 
 * Copyright (C) 2004-2009 Olivier Rolland <billl@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "shrip-help.h"
#include "shrip-common.h"

#include <signal.h>
#include <stdlib.h>
#include <unistd.h>

#include <ogmjob.h>

#include <glib/gi18n.h>
#include <glib/gstdio.h>

#ifdef HAVE_SYS_STATVFS_H
#include <sys/statvfs.h>
#endif

static gboolean     help          = FALSE;
static gboolean     debug         = FALSE;

extern gchar       *device;
extern const gchar *tmp_dir;

/*
 * Container options
 */
extern GType        format;
extern gint         tsize;
extern gint         tnumber;

/*
 * Misc options
 */
extern gboolean     ensure_sync;

/*
 * Video options
 */
extern GType        vcodec;
extern gint         passes;
extern guint        video_quality;
extern gint         scaler;
extern gint         threads;
extern gboolean     denoise;
extern gboolean     trellis;
extern gboolean     qpel;
extern gboolean     deblock;
extern gboolean     dering;
extern gboolean     turbo;

/*
 * Audio options
 */
extern GType        acodec;
extern gint         audio_quality;
extern gint         audio_srate;
extern gboolean     normalize;

/*
 * Subp options
 */
extern GType        scodec;
extern gint         charset;
extern gint         end_of_line;
extern gboolean     forced_sub;

/*
 * Advanced options
 */
extern gboolean     copy;
extern gboolean     keep;

/*
 * Other options
 */
extern guint        bitrate;
extern guint        crop_x;
extern guint        crop_y;
extern guint        crop_w;
extern guint        crop_h;
extern gdouble      bpp;
extern guint        scale_w;
extern guint        scale_h;
extern gint         deint;

/*
 * Selection
 */
extern gint         vid;
extern gboolean     aid[100];
extern gboolean     sid[100];
extern GSList      *alang;
extern GSList      *slang;
extern gboolean     relative;
extern gint         start_chap;
extern gint         end_chap;

static const GOptionEntry main_entries[] = 
{
  { "device",         0,  0, G_OPTION_ARG_CALLBACK, parse_device,    N_("Specify the DVD device (default: /dev/dvd)"),              "<device>"        },
  { "title",          0,  0, G_OPTION_ARG_INT,      &vid,            N_("Select the title to encode (default: longest)"),           "<id>",           },
  { "audio",          0,  0, G_OPTION_ARG_CALLBACK, parse_aid,       N_("Select the audio id to encode (default: none)"),           "<id>[,<id>]*"    },
  { "sub",            0,  0, G_OPTION_ARG_CALLBACK, parse_sid,       N_("Select the subtitle id to extract (default: none)"),       "<id>[,<id>]*"    },
  { "chapters",       0,  0, G_OPTION_ARG_CALLBACK, parse_chap,      N_("Specify which chapters to extract. (default: all)"),       "<start>[-<end>]" },
  { "container",      0,  0, G_OPTION_ARG_CALLBACK, parse_container, N_("Set the container to merge streams in (default: ogm)"),    "<container>"     },
  { "target-size",    0,  0, G_OPTION_ARG_INT,      &tsize,          N_("Set size of each target (default: 700MB)"),                "<size>"          },
  { "target-number",  0,  0, G_OPTION_ARG_INT,      &tnumber,        N_("Set the number of targets (default: 1)"),                  "<num>"           },
  { "fourcc",         0,  0, G_OPTION_ARG_CALLBACK, parse_fourcc,    N_("Set the video fourcc of the output file (default: auto)"), "<fourcc>"        },
  { "tmp-dir",        0,  0, G_OPTION_ARG_FILENAME, &tmp_dir,        N_("Set the temporary folder (default: /tmp)"),                "<directory>"     },
  { "relative",       0,  0, G_OPTION_ARG_NONE,     &relative,       N_("Extract selected chapters in relative mode"),              NULL              },
  { "ensure-sync",    0,  0, G_OPTION_ARG_NONE,     &ensure_sync,    N_("Ensure A/V synchronisation"),                              NULL              },
  { "copy",           0,  0, G_OPTION_ARG_NONE,     &copy,           N_("Copy DVD on hard drive before encoding"),                  NULL              },
  { "keep",           0,  0, G_OPTION_ARG_NONE,     &keep,           N_("Do not remove temporary files"),                           NULL              },
  { NULL,             0,  0, 0,                     NULL,            NULL,                                                          NULL              }
};

static const GOptionEntry help_entries[] =
{
  { "help", 'h', 0, G_OPTION_ARG_NONE, &help,  N_("Show help options"),     NULL },
  { "debug", 0,  0, G_OPTION_ARG_NONE, &debug, N_("Enable debug messages"), NULL },
  { NULL,    0,  0, 0,                 NULL,   NULL,                        NULL }
};

static const GOptionEntry video_entries[] =
{
  { "video-codec",   0,  0, G_OPTION_ARG_CALLBACK,  &parse_vcodec,   N_("Set the video codec (default: xvid)"),                      "<codec>"   },
  { "preset",        0,  0, G_OPTION_ARG_CALLBACK,  &parse_preset,   N_("Set video encoding quality (default: high)"),               "<preset>"  },
  { "passes",        0,  0, G_OPTION_ARG_INT,       &passes,         N_("Set the number of passes (default: 2)"),                    "<1-2>"     },
  { "threads",       0,  0, G_OPTION_ARG_INT,       &threads,        N_("Set the number of threads (default: 1)"),                   "<num>"     },
  { "scaler",        0,  0, G_OPTION_ARG_CALLBACK,  parse_scaler,    N_("Specify the software scaler (default: 7 - gauss)"),         "<0-10>"    },
  { "deinterlacer",  0,  0, G_OPTION_ARG_CALLBACK,  parse_deint,     N_("Specify the deinterlacer (default: 0 - none)"),             "<0-6>"     },
  { "bitrate",       0,  0, G_OPTION_ARG_INT,       &bitrate,        N_("Sets the bitrate to be used in kbps (default: auto)"),      "<num>"     },
  { "crop",          0,  0, G_OPTION_ARG_CALLBACK,  parse_crop,      N_("Crops the given part of the image (default: auto)"),        "<x:y:w:h>" },
  { "scale",         0,  0, G_OPTION_ARG_CALLBACK,  parse_scale,     N_("Scales the image to the given size (default: auto)"),       "<w:h>"     },
  { "denoise",       0,  0, G_OPTION_ARG_NONE,      &denoise,        N_("Reduce image noise"),                                       NULL        },
  { "trellis",       0,  0, G_OPTION_ARG_NONE,      &trellis,        N_("Trellis search quantization"),                              NULL        },
  { "qpel",          0,  0, G_OPTION_ARG_NONE,      &qpel,           N_("Use quarter pel motion compensation"),                      NULL        },
  { "turbo",         0,  0, G_OPTION_ARG_NONE,      &turbo,          N_("Speeds up pass one in 2-pass mode"),                        NULL        },
  { "deblock",       0,  0, G_OPTION_ARG_NONE,      &deblock,        N_("Use a deblocking filter"),                                  NULL        },
  { "dering",        0,  0, G_OPTION_ARG_NONE,      &dering,         N_("Use a deringing filter"),                                   NULL        },
  { "bpp",           0,  0, G_OPTION_ARG_CALLBACK,  parse_bpp,       N_("Sets the number of bits per pixel (default: 0.25)"),        "<num>"     },
  { NULL,            0,  0, 0,                      NULL,            NULL,                                                           NULL        }
};

static const GOptionEntry audio_entries[] =
{
  { "audio-codec",   0,  0, G_OPTION_ARG_CALLBACK, &parse_acodec,  N_("Set the audio codec (default: vorbis)"),            "<codec>" },
  { "audio-quality", 0,  0, G_OPTION_ARG_INT,      &audio_quality, N_("Set audio encoding quality (default: 3)"),          "<0-10>"  },
  { "audio-srate",   0,  0, G_OPTION_ARG_INT,      &audio_srate,   N_("Set the audio sample rate (default: 48000 Hz)"),    "<Hz>"    },
  { "normalize",     0,  0, G_OPTION_ARG_NONE,     &normalize,     N_("Maximize the volume without distorting the sound"), NULL      },
  { NULL,            0,  0, 0,                     NULL,           NULL,                                                   NULL      }
};

static const GOptionEntry subp_entries[] =
{
  { "subp-codec",  0,  0, G_OPTION_ARG_CALLBACK, &parse_scodec,  N_("Set the subtitle codec (default: vobsub)"),                 "<codec>"       },
  { "forced-sub",  0,  0, G_OPTION_ARG_NONE,     &forced_sub,    N_("Encode only forced subtitles"),                             NULL            },
  { "charset",     0,  0, G_OPTION_ARG_CALLBACK, &parse_charset, N_("Set the character set for text subtitles (default: utf8)"), "<charset>"     },
  { "newline",     0,  0, G_OPTION_ARG_CALLBACK, &parse_newline, N_("Set the newline code (default: lf)"),                       "<lf|cr|cr+lf>" },
  { NULL,          0,  0, 0,                     NULL,           NULL,                                                           NULL            }
};

static void
load_defaults (const gchar *filename)
{
  GKeyFile *key_file;

  key_file = g_key_file_new ();
  if (g_key_file_load_from_file (key_file, filename, G_KEY_FILE_NONE, NULL))
  {

    load_group_defaults (key_file, main_entries, "main");
    load_group_defaults (key_file, video_entries, "video");
    load_group_defaults (key_file, audio_entries, "audio");
    load_group_defaults (key_file, subp_entries, "sub");
  }
  g_key_file_free (key_file);
}

static void
init (gint *argc, gchar ***argv)
{
  (*argv)[0] = "shrip ccheck";

  signal (SIGTERM, sighandler); // kill
  signal (SIGHUP,  sighandler); // kill -HUP  /  xterm closed
  signal (SIGINT,  sighandler); // Interrupt from keyboard
  signal (SIGQUIT, sighandler); // Quit from keyboard
  signal (SIGPIPE, sighandler); // Some window managers cause this

  g_type_init ();

  ogmrip_plugin_init ();
}

static void
uninit (void)
{
  ogmrip_plugin_uninit ();
}

static void
usage (void)
{
  print_help ("ccheck",
      "<OUTPUT>",
      _("Help Options"),     help_entries, 
      _("Encode Options"),   main_entries, 
      _("Video Options"),    video_entries, 
      _("Audio Options"),    audio_entries, 
      _("Subtitle Options"), subp_entries, 
      NULL);

  g_print ("%s\n", _("See 'shrip encode --container help' for a list of available containers."));
  g_print ("%s\n", _("See 'shrip encode --video-codec help' for a list of available video codecs."));
  g_print ("%s\n", _("See 'shrip encode --audio-codec help' for a list of available audio codecs."));
  g_print ("%s\n", _("See 'shrip encode --sub-codec help' for a list of available subtitle codecs."));
  g_print ("%s\n", _("See 'shrip encode --video-quality help' for a list of available video quality presets."));
  g_print ("%s\n", _("See 'shrip encode --deinterlacer help' for a list of available deinterlacers."));
  g_print ("%s\n", _("See 'shrip encode --scaler help' for a list of available software scalers."));
  g_print ("%s\n", _("See 'shrip encode --charset help' for a list of available character sets."));
  g_print ("%s\n", _("See 'shrip encode --end-of-line help' for a list of available end of line styles."));
/*
  g_print ("%s\n", _("See 'shrip encode --fourcc help' for a list of available fourcc."));
*/
  g_print ("\n");
}

gint
ccheck (gint *argc, gchar ***argv)
{
  GError *error = NULL;
  gint result = OGMJOB_RESULT_ERROR;

  GOptionContext *context;
  GOptionGroup   *group;

  OGMRipEncoding    *encoding = NULL;

  OGMDvdDisc        *disc  = NULL;
  OGMDvdTitle       *title = NULL;

  gchar *filename = NULL;

  init (argc, argv);

  filename = g_build_filename (SHRIP_CONF_DIR, "shrip.conf", NULL);
  load_defaults (filename);
  g_free (filename);

  filename = g_build_filename (g_get_home_dir (), ".shrip", NULL);
  load_defaults (filename);
  g_free (filename);

  filename = NULL;

  context = g_option_context_new ("<OUTPUT>");
  g_option_context_set_help_enabled (context, FALSE);
  g_option_context_add_main_entries (context, main_entries, NULL);

  group = g_option_group_new ("help", _("Help Options"), _("Show help options"), NULL, NULL);
  g_option_group_add_entries (group, help_entries);
  g_option_context_add_group (context, group);

  group = g_option_group_new ("video", _("Video Options"), _("Show video help options"), NULL, NULL);
  g_option_group_add_entries (group, video_entries);
  g_option_context_add_group (context, group);

  group = g_option_group_new ("audio", _("Audio Options"), _("Show audio help options"), NULL, NULL);
  g_option_group_add_entries (group, audio_entries);
  g_option_context_add_group (context, group);

  group = g_option_group_new ("subtitle", _("Subtitle Options"), _("Show subtitle help options"), NULL, NULL);
  g_option_group_add_entries (group, subp_entries);
  g_option_context_add_group (context, group);

  if (!g_option_context_parse (context, argc, argv, &error))
  {
    g_option_context_free (context);
    goto cleanup;
  }
  g_option_context_free (context);

  if (tmp_dir && g_file_test (tmp_dir, G_FILE_TEST_IS_DIR))
    ogmrip_fs_set_tmp_dir (tmp_dir);
  else
    ogmrip_fs_set_tmp_dir (g_get_tmp_dir ());

  if (help)
  {
    usage ();
    return EXIT_SUCCESS;
  }

  if (debug)
    ogmjob_log_set_print_stdout (TRUE);

  disc = ogmdvd_disc_new (device, &error);
  if (!disc || !ogmdvd_disc_open (disc, &error))
    goto cleanup;

  if (vid == 0)
    vid = get_longest_title (disc);

  if (vid < 1 || vid > ogmdvd_disc_get_n_titles (disc))
    goto cleanup;

  title = ogmdvd_disc_get_nth_title (disc, vid - 1);
  ogmdvd_disc_unref (disc);

  if (!title)
    goto cleanup;

  encoding = create_encoding (title, &error);
  if (!encoding)
    goto cleanup;

  result = ogmrip_encoding_test (encoding, &error);

cleanup:
  g_free (device);

  if (title)
    ogmdvd_title_unref (title);

  if (encoding)
    g_object_unref (encoding);

  g_slist_free (alang);
  g_slist_free (slang);

  if (error)
  {
    g_printerr ("%s\n", error->message);
    g_error_free (error);
  }

  uninit ();

  if (result != OGMJOB_RESULT_SUCCESS)
    return EXIT_FAILURE;

  return EXIT_SUCCESS;
}

