/*
 * shrip - A DVD Encoder 
 * Copyright (C) 2004-2009 Olivier Rolland <billl@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "shrip-common.h"

#include <ogmjob.h>

#include <math.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>

#include <glib/gi18n.h>
#include <glib/gstdio.h>

gchar       *device        = NULL;
const gchar *output        = NULL;
const gchar *tmp_dir       = NULL;
const gchar *label         = NULL;
const gchar *profile       = NULL;

/*
 * Container options
 */

GType        format        = G_TYPE_NONE;
const gchar *fourcc        = NULL;
gint         tsize         = 700;
gint         tnumber       = 1;
gint         chap_lang     = 0;

/*
 * Misc options
 */
gboolean     ensure_sync   = FALSE;
gboolean     test          = FALSE;

/*
 * Video options
 */
GType        vcodec        = G_TYPE_NONE;
gint         passes        = 2;
guint        preset        = OGMRIP_QUALITY_HIGH;
gint         scaler        = OGMRIP_SCALER_GAUSS;
gboolean     denoise       = FALSE;
gboolean     trellis       = FALSE;
gboolean     qpel          = FALSE;
gboolean     deblock       = FALSE;
gboolean     dering        = FALSE;
gboolean     turbo         = FALSE;
gboolean     cartoon       = FALSE;
gboolean     can_crop      = TRUE;
gboolean     can_scale     = TRUE;

#ifdef HAVE_SYSCONF_NPROC
gint         threads       = 0;
#else
gint         threads       = 1;
#endif

/*
 * Audio options
 */
GType        acodec        = G_TYPE_NONE;
gint         audio_quality = 3;
gint         audio_srate   = 48000;
gboolean     normalize     = FALSE;

/*
 * Subp options
 */
GType        scodec        = G_TYPE_NONE;
gint         charset       = 0;
gint         newline       = 0;
gboolean     forced_sub    = FALSE;

/*
 * Advanced options
 */
gboolean     copy          = FALSE;
gboolean     keep          = FALSE;

/*
 * Other options
 */
guint        method        = OGMRIP_ENCODING_SIZE;
guint        bitrate       = 0;
gdouble      quantizer     = 0;
guint        crop          = OGMRIP_OPTIONS_AUTOMATIC;
guint        crop_x        = 0;
guint        crop_y        = 0;
guint        crop_w        = 0;
guint        crop_h        = 0;
gdouble      bpp           = 0.25; 
guint        scale         = OGMRIP_OPTIONS_AUTOMATIC;
guint        scale_w       = 0;
guint        scale_h       = 0;
gint         deint         = OGMRIP_DEINT_NONE;
guint        max_w         = 0;
guint        max_h         = 0;
gboolean     expand        = FALSE;

/*
 * Selection
 */
gint         vid           = 0;
gboolean     aid[100];
gboolean     sid[100];
GSList      *alang         = NULL;
GSList      *slang         = NULL;
gboolean     relative      = FALSE;
gint         start_chap    = 0;
gint         end_chap      = -1;

OGMRipEdl   *edl           = NULL;

/*
 * Global variables
 */

static OGMRipEncoding *encoding = NULL;

static gint         naid  = 0;
static gint         nsid  = 0;

static gulong       start = 0;

gboolean
parse_device (const gchar *name, const gchar *value, gpointer data, GError **error)
{
  struct stat buf;

  if (g_stat (value, &buf) < 0)
  {
    g_set_error (error, G_FILE_ERROR, g_file_error_from_errno (errno),
        _("Failed to stat '%s': %s"), value, g_strerror (errno));
    return FALSE;
  }

  device = ogmrip_fs_get_full_path (value);

  return TRUE;
}

gboolean
parse_chap (const gchar *name, const gchar *value, gpointer data, GError **error)
{
  gchar *str;

  if (!value || strlen (value) == 0)
    return FALSE;

  errno = 0;
  start_chap = strtoul (value, &str, 10);

  if (!errno && *str == '-')
    end_chap = strtoul (str + 1, &str, 10);

  if (errno == ERANGE || *str != '\0')
  {
    g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
        _("Cannot parse value '%s' for %s"), value, name);
    return FALSE;
  }

  if (end_chap != -1 && start_chap > end_chap)
  {
    gint swap;

    swap = start_chap;
    start_chap = end_chap;
    end_chap = swap;
  }

  return TRUE;
}

static gboolean
parse_alang (const gchar *str, gchar **endptr, GError **error)
{
  guint code;

  if (str[0] == '\0' || str[1] == '\0' || (str[2] != ',' && str[2] != '\0'))
    return FALSE;

  code = (str[0] << 8) | str[1];
  alang = g_slist_append (alang, GUINT_TO_POINTER (code));

  if (endptr)
    *endptr = (gchar *) str + 2;

  return TRUE;
}

gboolean
parse_aid (const gchar *name, const gchar *value, gpointer data, GError **error)
{
  gchar *str = (gchar *) value, *endptr;
  guint id;

  if (!value || strlen (value) == 0)
    return FALSE;

  while (TRUE)
  {
    errno = 0;
    endptr = NULL;

    id = strtoul (str, &endptr, 10);

    if (errno == 0)
    {
      if (id > 0 && id < 100)
      {
        if (!aid[id - 1])
          naid ++;

        aid[id - 1] = TRUE;
      }
      else
      {
        errno = 1;
        if (str == endptr && parse_alang (str, &endptr, error))
          errno = 0;
      }

      if (*endptr != '\0' && *endptr != ',')
        errno = 1;
    }

    if (errno)
    {
      g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
          _("Cannot parse value '%s' for %s"), value, name);
      return FALSE;
    }

    str = endptr;

    if (*str == '\0')
      break;
    str ++;
  } 

  return TRUE;
}

static void
convert_alang (OGMDvdTitle *title)
{
  if (alang)
  {
    GSList *streams, *stream;
    GSList *link;
    guint code, id;

    streams = ogmdvd_title_get_audio_streams (title);
    if (streams)
    {
      for (link = alang; link; link = link->next)
      {
        code = GPOINTER_TO_UINT (link->data);

        for (stream = streams; stream; stream = stream->next)
        {
          if (ogmdvd_audio_stream_get_language (OGMDVD_AUDIO_STREAM (stream->data)) == code)
          {
            id = ogmdvd_stream_get_nr (OGMDVD_STREAM (stream->data));

            if (!aid[id])
              naid ++;

            aid[id] = TRUE;

            break;
          }
        }
      }

      g_slist_foreach (streams, (GFunc) ogmdvd_stream_unref, NULL);
      g_slist_free (streams);
    }
  }
}

static gboolean
parse_slang (const gchar *str, gchar **endptr, GError **error)
{
  guint code;

  if (str[0] == '\0' || str[1] == '\0' || (str[2] != ',' && str[2] != '\0'))
    return FALSE;

  code = (str[0] << 8) | str[1];
  slang = g_slist_append (slang, GUINT_TO_POINTER (code));

  if (endptr)
    *endptr = (gchar *) str + 2;

  return TRUE;
}

gboolean
parse_sid (const gchar *name, const gchar *value, gpointer data, GError **error)
{
  gchar *str = (gchar *) value, *endptr;
  guint id;

  if (!value || strlen (value) == 0)
    return FALSE;

  while (TRUE)
  {
    errno = 0;
    endptr = NULL;

    id = strtoul (str, &endptr, 10);

    if (errno == 0)
    {
      if (id > 0 && id < 100)
      {
        if (!sid[id - 1])
          nsid ++;

        sid[id - 1] = TRUE;
      }
      else
      {
        errno = 1;
        if (str == endptr && parse_slang (str, &endptr, error))
          errno = 0;
      }

      if (*endptr != '\0' && *endptr != ',')
        errno = 1;
    }

    if (errno)
    {
      g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
          _("Cannot parse value '%s' for %s"), value, name);
      return FALSE;
    }

    str = endptr;

    if (*str == '\0')
      break;
    str ++;
  } 

  return TRUE;
}

static void
convert_slang (OGMDvdTitle *title)
{
  if (slang)
  {
    GSList *streams, *stream;
    GSList *link;
    guint code, id;

    streams = ogmdvd_title_get_subp_streams (title);
    if (streams)
    {
      for (link = alang; link; link = link->next)
      {
        code = GPOINTER_TO_UINT (link->data);

        for (stream = streams; stream; stream = stream->next)
        {
          if (ogmdvd_subp_stream_get_language (OGMDVD_SUBP_STREAM (stream->data)) == code)
          {
            id = ogmdvd_stream_get_nr (OGMDVD_STREAM (stream->data));

            if (!sid[id])
              nsid ++;

            sid[id] = TRUE;

            break;
          }
        }
      }

      g_slist_foreach (streams, (GFunc) ogmdvd_stream_unref, NULL);
      g_slist_free (streams);
    }
  }
}

void
print_codec (GType type, const gchar *name, const gchar *description, gpointer data)
{
  g_print ("  %6s - %s\n", name, description);
}

static gint
find_codec (GType type, const gchar *name1, const gchar *description, const gchar *name2)
{
  return strcmp (name1, name2);
}

gboolean
parse_container (const gchar *name, const gchar *value, gpointer data, GError **error)
{
  if (!value || strlen (value) == 0)
    return FALSE;

  if (g_str_equal (value, "help"))
  {
    g_print ("%s:\n", _("Available containers"));
    ogmrip_plugin_foreach_container (print_codec, NULL);

    return TRUE;
  }

  format = ogmrip_plugin_find_container ((OGMRipPluginCmpFunc) find_codec, value);
  if (format == G_TYPE_NONE)
  {
    g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
        _("The container '%s' does not exist."), value);
    return FALSE;
  }

  return TRUE;
}

gboolean
parse_vcodec (const gchar *name, const gchar *value, gpointer data, GError **error)
{
  if (!value || strlen (value) == 0)
    return FALSE;

  vcodec = ogmrip_plugin_find_video_codec ((OGMRipPluginCmpFunc) find_codec, value);
  if (vcodec == G_TYPE_NONE)
  {
    if (g_str_equal (value, "help"))
    {
      g_print ("%s:\n", _("Available video codecs"));
      ogmrip_plugin_foreach_video_codec (print_codec, NULL);
    }
    else
    {
      g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
          _("The video codec '%s' does not exist."), value);
    }
    return FALSE;
  }

  return TRUE;
}

gboolean
parse_preset (const gchar *name, const gchar *value, gpointer data, GError **error)
{
  if (!value || strlen (value) == 0)
    return FALSE;

  if (g_str_equal (value, "normal"))
    preset = OGMRIP_QUALITY_NORMAL;
  else if (g_str_equal (value, "high"))
    preset = OGMRIP_QUALITY_HIGH;
  else if (g_str_equal (value, "extreme"))
    preset = OGMRIP_QUALITY_EXTREME;
  else
  {
    if (g_str_equal (value, "help"))
    {
      g_print ("%s:\n", _("Available video quality presets"));
      g_print ("   normal - %s\n", _("Normal quality"));
      g_print ("     high - %s\n", _("High quality"));
      g_print ("  extreme - %s\n", _("Very high quality"));
    }
    else
    {
      g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
          _("The video quality preset '%s' does not exist."), value);
    }
    return FALSE;
  }

  return TRUE;
}

gboolean
parse_acodec (const gchar *name, const gchar *value, gpointer data, GError **error)
{
  if (!value || strlen (value) == 0)
    return FALSE;

  acodec = ogmrip_plugin_find_audio_codec ((OGMRipPluginCmpFunc) find_codec, value);
  if (acodec == G_TYPE_NONE)
  {
    if (g_str_equal (value, "help"))
    {
      g_print ("%s:\n", _("Available audio codecs"));
      ogmrip_plugin_foreach_audio_codec (print_codec, NULL);
    }
    else
    {
      g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
          _("The audio codec '%s' does not exist."), value);
    }
    return FALSE;
  }

  return TRUE;
}

gboolean
parse_scodec (const gchar *name, const gchar *value, gpointer data, GError **error)
{
  if (!value || strlen (value) == 0)
    return FALSE;

  scodec = ogmrip_plugin_find_subp_codec ((OGMRipPluginCmpFunc) find_codec, value);
  if (scodec == G_TYPE_NONE)
  {
    if (g_str_equal (value, "help"))
    {
      g_print ("%s:\n", _("Available subtitles codecs"));
      ogmrip_plugin_foreach_subp_codec (print_codec, NULL);
    }
    else
    {
      g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
          _("The subtitle codec '%s' does not exist."), value);
    }
    return FALSE;
  }

  return TRUE;
}

gboolean
parse_scaler (const gchar *name, const gchar *value, gpointer data, GError **error)
{
  gchar *end;

  if (!value)
    return FALSE;

  if (g_str_equal (value, "help"))
  {
    g_print ("%s:\n", _("Available software scalers"));
    g_print ("   0 - %s\n", _("Fast bilinear"));
    g_print ("   1 - %s\n", _("Bilinear"));
    g_print ("   2 - %s\n", _("Bicubic"));
    g_print ("   3 - %s\n", _("Experimental"));
    g_print ("   4 - %s\n", _("Nearest neighbour"));
    g_print ("   5 - %s\n", _("Area"));
    g_print ("   6 - %s\n", _("Luma bicubic / chroma bilinear"));
    g_print ("   7 - %s\n", _("Gauss (best for downscaling, default)"));
    g_print ("   8 - %s\n", _("SincR"));
    g_print ("   9 - %s\n", _("Lanczos (best for upscaling)"));
    g_print ("  10 - %s\n", _("Natural bicubic spline"));

    return FALSE;
  }

  errno = 0;
  scaler = strtol (value, &end, 0);

  if (*value == '\0' || *end != '\0' || errno == ERANGE)
  {
    g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
        _("Cannot parse integer value '%s' for %s"), value, name);
    return FALSE;
  }

  if (scaler <= OGMRIP_SCALER_FAST_BILINEAR || scaler >= OGMRIP_SCALER_BICUBIC_SPLINE)
  {
    g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
        _("The scaler %d does not exist."), scaler);
    return FALSE;
  }

  return TRUE;
}

gboolean
parse_deint (const gchar *name, const gchar *value, gpointer data, GError **error)
{
  gchar *end;

  if (!value)
    return FALSE;

  if (g_str_equal (value, "help"))
  {
    g_print ("%s:\n", _("Available deinterlacers"));
    g_print ("  0 - %s\n", _("None (default)"));
    g_print ("  1 - %s\n", _("Linear blend deinterlacer"));
    g_print ("  2 - %s\n", _("Linear interpolating deinterlacer"));
    g_print ("  3 - %s\n", _("Cubic interpolating deinterlacer"));
    g_print ("  4 - %s\n", _("Median deinterlacer"));
    g_print ("  5 - %s\n", _("FFMpeg deinterlacer"));
    g_print ("  6 - %s\n", _("Adaptive kernel deinterlacer"));

    return FALSE;
  }

  errno = 0;
  deint = strtol (value, &end, 0);

  if (*value == '\0' || *end != '\0' || errno == ERANGE)
  {
    g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
        _("Cannot parse integer value '%s' for %s"), value, name);
    return FALSE;
  }

  if (deint <= OGMRIP_DEINT_NONE || deint >= OGMRIP_DEINT_KERNEL)
  {
    g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
        _("The deinterlacer %d does not exist."), deint);
    return FALSE;
  }

  return TRUE;
}

gboolean
parse_fourcc (const gchar *name, const gchar *value, gpointer data, GError **error)
{
  if (!value || strlen (value) == 0)
    return FALSE;

  if (g_str_equal (value, "auto"))
    fourcc = NULL;
  else if (g_str_equal (value, "xvid"))
    fourcc = "XVID";
  else if (g_str_equal (value, "divx4"))
    fourcc = "DIVX";
  else if (g_str_equal (value, "divx5"))
    fourcc = "DX50";
  else if (g_str_equal (value, "ffmpeg"))
    fourcc = "FMP4";
  else
  {
    if (g_str_equal (value, "help"))
    {
      g_print ("%s:\n", _("Available fourcc"));
      g_print ("    auto - %s\n", _("Automatic"));
      g_print ("    xvid - %s\n", _("XviD"));
      g_print ("   divx4 - %s\n", _("DivX 4"));
      g_print ("   divx5 - %s\n", _("DivX 5"));
      g_print ("  ffmpeg - %s\n", _("FFMpeg"));
    }
    else
    {
      g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
          _("The fourcc '%s' does not exist."), value);
    }
    return FALSE;
  }

  return TRUE;
}

gboolean
parse_crop (const gchar *name, const gchar *value, gpointer data, GError **error)
{
  gchar *str;

  if (!value)
    return FALSE;

  if (g_str_equal (value, "auto"))
  {
    crop = OGMRIP_OPTIONS_AUTOMATIC;
    return TRUE;
  }

  if (g_str_equal (value, "none"))
  {
    crop = OGMRIP_OPTIONS_NONE;
    return TRUE;
  }

  errno = 0;
  crop_x = strtoul (value, &str, 10);
  if (str[0] != ':' || str[1] == '\0' || errno == ERANGE)
  {
    g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
        _("Cannot parse value '%s' for %s"), value, name);
    return FALSE;
  }

  errno = 0;
  crop_y = strtoul (str + 1, &str, 10);
  if (str[0] != ':' || str[1] == '\0' || errno == ERANGE)
  {
    g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
        _("Cannot parse value '%s' for %s"), value, name);
    return FALSE;
  }

  errno = 0;
  crop_w = strtoul (str + 1, &str, 10);
  if (str[0] != ':' || str[1] == '\0' || errno == ERANGE)
  {
    g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
        _("Cannot parse value '%s' for %s"), value, name);
    return FALSE;
  }

  errno = 0;
  crop_h = strtoul (str + 1, &str, 10);
  if (str[0] != '\0' || errno == ERANGE)
  {
    g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
        _("Cannot parse value '%s' for %s"), value, name);
    return FALSE;
  }

  crop = OGMRIP_OPTIONS_MANUAL;

  return TRUE;
}

gboolean
parse_scale (const gchar *name, const gchar *value, gpointer data, GError **error)
{
  gchar *str;

  if (!value)
    return FALSE;

  if (g_str_equal (value, "auto"))
  {
    scale = OGMRIP_OPTIONS_AUTOMATIC;
    return TRUE;
  }

  if (g_str_equal (value, "none"))
  {
    scale = OGMRIP_OPTIONS_NONE;
    return TRUE;
  }

  errno = 0;
  scale_w = strtoul (value, &str, 10);
  if (str[0] != ':' || str[1] == '\0' || errno == ERANGE)
  {
    g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
        _("Cannot parse value '%s' for %s"), value, name);
    return FALSE;
  }

  errno = 0;
  scale_h = strtoul (str + 1, &str, 10);
  if (*str != '\0' || errno == ERANGE)
  {
    g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
        _("Cannot parse value '%s' for %s"), value, name);
    return FALSE;
  }

  scale = OGMRIP_OPTIONS_MANUAL;

  return TRUE;
}

gboolean
parse_size (const gchar *name, const gchar *value, gpointer data, GError **error)
{
  gchar *str;

  if (!value)
    return FALSE;

  if (g_str_equal (value, "none"))
  {
    max_w = max_h = 0;
    expand = FALSE;

    return TRUE;
  }

  errno = 0;
  max_w = strtoul (value, &str, 10);
  if (str[0] != ':' || str[1] == '\0' || errno == ERANGE)
  {
    g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
        _("Cannot parse value '%s' for %s"), value, name);
    return FALSE;
  }

  errno = 0;
  max_h = strtoul (str + 1, &str, 10);
  if ((*str != '\0' && *str != ':') || errno == ERANGE)
  {
    g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
        _("Cannot parse value '%s' for %s"), value, name);
    return FALSE;
  }

  if (*str == ':')
  {
    if (g_str_equal (str + 1, "true"))
      expand = TRUE;
    else if (g_str_equal (str + 1, "false"))
      expand = FALSE;
    else
    {
      g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
          _("Cannot parse value '%s' for %s"), value, name);
      return FALSE;
    }
  }

  return TRUE;
}

gboolean
parse_charset (const gchar *name, const gchar *value, gpointer data, GError **error)
{
  if (!value || strlen (value) == 0)
    return FALSE;

  if (g_str_equal (value, "utf8"))
    charset = OGMRIP_CHARSET_UTF8;
  else if (g_str_equal (value, "iso8859-1"))
    charset = OGMRIP_CHARSET_ISO8859_1;
  else if (g_str_equal (value, "ascii"))
    charset = OGMRIP_CHARSET_ASCII;
  else
  {
    if (g_str_equal (value, "help"))
    {
      g_print ("%s:\n", _("Available character sets"));
      g_print ("       utf8 - %s\n", _("UTF-8"));
      g_print ("  iso8859-1 - %s\n", _("ISO-8859-1"));
      g_print ("      ascii - %s\n", _("ASCII"));
    }
    else
    {
      g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
          _("The character set '%s' does not exist."), value);
    }
    return FALSE;
  }

  return TRUE;
}

gboolean
parse_newline (const gchar *name, const gchar *value, gpointer data, GError **error)
{
  if (!value || strlen (value) == 0)
    return FALSE;

  if (g_str_equal (value, "lf"))
    newline = OGMRIP_NEWLINE_LF;
  else if (g_str_equal (value, "cr+lf"))
    newline = OGMRIP_NEWLINE_CR_LF;
  else if (g_str_equal (value, "cr"))
    newline = OGMRIP_NEWLINE_CR;
  else
  {
    if (g_str_equal (value, "help"))
    {
      g_print ("%s:\n", _("Available end of line styles"));
      g_print ("  unix - %s\n", _("Carriage return only"));
      g_print ("   dos - %s\n", _("Carriage return and line feed"));
    }
    else
    {
      g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, 
          _("The newline code '%s' does not exist."), value);
    }
    return FALSE;
  }

  return TRUE;
}

gboolean
parse_bpp (const gchar *name, const gchar *value, gpointer data, GError **error)
{
  if (!value || strlen (value) == 0)
    return FALSE;

  errno = 0;
  bpp = strtod (value, NULL);
  if (errno)
  {
    g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
        _("Cannot parse value '%s' for %s"), value, name);
    return FALSE;
  }

  if (bpp <= 0.0 || bpp > 1.0)
  {
    g_set_error (error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE,
        _("The number of bits per pixel must be greater than 0 and lesser or equal to 1."));
    return FALSE;
  }

  return TRUE;
}

gboolean
parse_profile (const gchar *name, const gchar *value, gpointer data, GError **error)
{
  GSList *profiles = NULL;

  if (!value || strlen (value) == 0)
    return FALSE;

  /* TODO */
  /* profiles = ogmrip_settings_get_subsections (settings, OGMRIP_GCONF_PROFILES); */

  if (g_str_equal (value, "help"))
    g_print ("%s:\n", _("Available profiles"));
  else
  {
    GSList *link;

    for (link = profiles; !profile && link; link = link->next)
    {
      if (g_str_equal (value, link->data))
        profile = value;
    }
  }

  g_slist_foreach (profiles, (GFunc) g_free, NULL);
  g_slist_free (profiles);

  return profile != NULL;
}

gint
get_longest_title (OGMDvdDisc *disc)
{
  GSList *titles;
  gdouble longest = 0, length;
  gint vid = 0;

  titles = ogmdvd_disc_get_titles (disc);
  if (titles)
  {
    GSList *title;

    for (title = titles; title; title = title->next)
    {
      length = ogmdvd_title_get_length (OGMDVD_TITLE (title->data), NULL);
      if (length > longest)
      {
        longest = length;
        vid = ogmdvd_title_get_nr (OGMDVD_TITLE (title->data)) + 1;
      }
    }

    g_slist_foreach (titles, (GFunc) ogmdvd_title_unref, NULL);
    g_slist_free (titles);
  }

  return vid;
}

static void
load_boolean_default (GKeyFile *key_file, const gchar *group_name, const gchar *key_name, gboolean *value)
{
  GError *error = NULL;
  gboolean val;

  val = g_key_file_get_boolean (key_file, group_name, key_name, &error);
  if (error)
    g_error_free (error);
  else
    *value = val;
}

static void
load_int_default (GKeyFile *key_file, const gchar *group_name, const gchar *key_name, gint *value)
{
  GError *error = NULL;
  gint val;

  val = g_key_file_get_integer (key_file, group_name, key_name, &error);
  if (error)
    g_error_free (error);
  else
    *value = val;
}

static void
load_filename_default (GKeyFile *key_file, const gchar *group_name, const gchar *key_name, gchar **value)
{
  GError *error = NULL;
  gchar *val;

  val = g_key_file_get_locale_string (key_file, group_name, key_name, "UTF8", &error);
  if (error)
    g_error_free (error);
  else
  {
    *value = g_filename_from_utf8 (val, -1, NULL, NULL, NULL);
    g_free (val);
  }

}

static void
load_callback_default (GKeyFile *key_file, const gchar *group_name, const gchar *key_name, GOptionArgFunc func)
{
  GError *error = NULL;
  gchar *val;

  val = g_key_file_get_string (key_file, group_name, key_name, &error);
  if (error)
    g_error_free (error);
  else
  {
    func (key_name, val, NULL, NULL);
    g_free (val);
  }
}

void
load_group_defaults (GKeyFile *key_file, const GOptionEntry *entries, const gchar *group_name)
{
  guint i;

  for (i = 0; entries[i].long_name; i++)
  {
    switch (entries[i].arg)
    {
      case G_OPTION_ARG_NONE:
        load_boolean_default (key_file, group_name, entries[i].long_name, (gboolean *) entries[i].arg_data);
        break;
      case G_OPTION_ARG_INT:
        load_int_default (key_file, group_name, entries[i].long_name, (gint *) entries[i].arg_data);
        break;
      case G_OPTION_ARG_CALLBACK:
        load_callback_default (key_file, group_name, entries[i].long_name, (GOptionArgFunc) entries[i].arg_data);
        break;
      case G_OPTION_ARG_FILENAME:
        load_filename_default (key_file, group_name, entries[i].long_name, (gchar **) entries[i].arg_data);
        break;
      default:
        break;
    }
  }
}

static gboolean
check_options (OGMDvdTitle *title)
{
  gint i, num_chap, num_streams;

  if (!device)
    device = g_strdup ("/dev/dvd");
  else
    device = g_strdup (device);

  if (tnumber < 1)
  {
    g_printerr ("%s\n", _("The target number must be positive."));
    return FALSE;
  }

  if (tsize < 1)
  {
    g_printerr ("%s\n", _("The size of the targets must be positive."));
    return FALSE;
  }

  if (passes < 1)
  {
    g_printerr ("%s\n", _("The pass number must be positive."));
    return FALSE;
  }

#ifndef HAVE_SYSCONF_NPROC
  if (threads < 1)
  {
    g_printerr ("%s\n", _("The number of threads must be positive."));
    return FALSE;
  }
#endif

  if (audio_quality < 0 || audio_quality > 10)
  {
    g_printerr ("%s\n", _("The audio quality must be between 0 and 10."));
    return FALSE;
  }

  if (audio_srate < 1)
  {
    g_printerr ("%s\n", _("The sample rate must be positive."));
    return FALSE;
  }

  /*
   * Check audio streams
   */
  num_streams = ogmdvd_title_get_n_audio_streams (title);
  for (i = 0; i < 100; i++)
  {
    if (aid[i] && i >= num_streams)
    {
      g_printerr (_("The audio stream %d does not exist."), i + 1);
      g_printerr ("\n");
      return FALSE;
    }
  }

  /*
   * Check subp streams
   */
  num_streams = ogmdvd_title_get_n_subp_streams (title);
  for (i = 0; i < 100; i++)
  {
    if (sid[i] && i >= num_streams)
    {
      g_printerr (_("The subtitles stream %d does not exist."), i + 1);
      g_printerr ("\n");
      return FALSE;
    }
  }

  /*
   * Check chapters
   */

  num_chap = ogmdvd_title_get_n_chapters (title);
  if (end_chap == -1)
    end_chap = num_chap;
  start_chap --;
  end_chap --;

  if (start_chap < 0)
  {
    g_printerr (_("The start chapter %d does not exist."), start_chap);
    g_printerr ("\n");
    return FALSE;
  }

  if (end_chap >= num_chap)
  {
    g_printerr (_("The last chapter %d does not exist."), end_chap);
    g_printerr ("\n");
    return FALSE;
  }

  if (quantizer > 0.0 && bitrate > 0)
  {
    g_printerr (_("Cannot set a bitrate and a quantizer at the same time."));
    g_printerr ("\n");
    return FALSE;
  }

  method = OGMRIP_ENCODING_SIZE;

  if (bitrate > 0)
    method = OGMRIP_ENCODING_BITRATE;

  if (quantizer > 0.0)
    method = OGMRIP_ENCODING_QUANTIZER;

  return TRUE;
}

void
sighandler (int sig)
{
  if (encoding)
    ogmrip_encoding_cancel (encoding);
}

static const gchar * const messages[] =
{
  N_("DVD Backup"),
  N_("Analyzing video stream"),
  N_("Extracting chapters information"),
  N_("Extracting audio stream"),
  N_("Extracting subtitles stream"),
  N_("Cropping"),
  N_("Compressibility Test"),
  N_("Encoding video title"),
  N_("Merging audio and video streams")
};

static void
encoding_run_callback (OGMRipEncoding *encoding, OGMRipEncodingTask *task)
{
  GTimeVal tv;

  g_get_current_time (&tv);
  start = tv.tv_sec;

  if (task->type == OGMRIP_TASK_CHAPTERS)
    g_print ("\r%s", gettext (messages[task->type]));
}

static void
encoding_progress_callback (OGMRipEncoding *encoding, OGMRipEncodingTask *task)
{
  static gint type = -1;
  static gchar *message = NULL;

  GTimeVal tv;
  gulong eta;

  if (type != task->type)
  {
    OGMDvdTitle *title;
    OGMDvdAudioStream *astream;
    OGMDvdSubpStream *sstream;

    if (message)
      g_free (message);

    type = task->type;
    switch (type)
    {
      case OGMRIP_TASK_VIDEO:
        title = ogmrip_codec_get_input (OGMRIP_CODEC (task->spawn));
        message = g_strdup_printf ("\r%s %d", gettext (messages[task->type]), ogmdvd_title_get_nr (title) + 1);
        break;
      case OGMRIP_TASK_AUDIO:
        astream = ogmrip_audio_codec_get_dvd_audio_stream (OGMRIP_AUDIO_CODEC (task->spawn));
        message = g_strdup_printf ("\r%s %d", gettext (messages[task->type]), ogmdvd_stream_get_nr (OGMDVD_STREAM (astream)) + 1);
        break;
      case OGMRIP_TASK_SUBP:
        sstream = ogmrip_subp_codec_get_dvd_subp_stream (OGMRIP_SUBP_CODEC (task->spawn));
        message = g_strdup_printf ("\r%s %d", gettext (messages[task->type]), ogmdvd_stream_get_nr (OGMDVD_STREAM (sstream)) + 1);
        break;
      default:
        message = g_strdup_printf ("\r%s", gettext (messages[task->type]));
        break;
    }
  }

  g_get_current_time (&tv);

  eta = (1.0 - task->detail.fraction) * (tv.tv_sec - start) / task->detail.fraction;

  g_print ("%s: %.0lf%% (eta: %02lu:%02lu:%02lu)", message,
      task->detail.fraction * 100, eta / 3600, (eta % 3600) / 60, eta % 60);
}

static void
encoding_complete_callback (OGMRipEncoding *encoding, OGMRipEncodingTask *task)
{
  g_print ("\n");
}

static void
encoding_callback (OGMRipEncoding *encoding, OGMRipEncodingTask *task)
{
  switch (task->event)
  {
    case OGMRIP_TASK_RUN:
      encoding_run_callback (encoding, task);
      break;
    case OGMRIP_TASK_PROGRESS:
      encoding_progress_callback (encoding, task);
      break;
    case OGMRIP_TASK_COMPLETE:
      encoding_complete_callback (encoding, task);
      break;
    default:
      break;
  }
}

OGMRipEncoding *
create_encoding (OGMDvdTitle *title, GError **error)
{
  OGMRipOptions options;

  OGMDvdAudioStream *astream;
  OGMDvdSubpStream *sstream;

  guint i;

  if (!check_options (title))
    return NULL;

  encoding = ogmrip_encoding_new (title, output);

  if (format == G_TYPE_NONE)
    format = ogmrip_plugin_get_nth_container (0);

  if (!ogmrip_encoding_set_container_type (encoding, format, error))
  {
    g_printerr (_("Cannot set container."));
    g_printerr ("\n");

    g_object_unref (encoding);
    return NULL;
  }

  if (vcodec == G_TYPE_NONE)
    vcodec = ogmrip_plugin_get_nth_video_codec (0);

  if (!ogmrip_encoding_set_video_codec_type (encoding, vcodec, error))
  {
    g_printerr (_("Cannot set video codec."));
    g_printerr ("\n");

    g_object_unref (encoding);
    return NULL;
  }

  if (profile)
    ogmrip_encoding_set_profile (encoding, profile);

  g_object_set (encoding,
      "bpp",            bpp,
      "can-crop",       can_crop,
      "can-scale",      can_scale,
      "cartoon",        cartoon,
      "chapters-lang",  chap_lang,
      "copy-dvd",       copy,
      "deblock",        deblock,
      "deinterlacer",   deint,
      "denoise",        denoise,
      "dering",         dering,
      "end-chapter",    end_chap,
      "ensure-sync",    ensure_sync,
      "expand",         expand,
      "keep-tmp-files", keep,
      "label",          label,
      "max-height",     max_h,
      "max-width",      max_w,
      "method",         method,
      "passes",         passes,
      "preset",         preset,
      "qpel",           qpel,
      "quantizer",      quantizer,
      "relative",       relative,
      "scaler",         scaler,
      "start-chapter",  start_chap,
      "target-number",  tnumber,
      "target-size",    tsize,
      "test",           test,
      "threads",        threads,
      "trellis",        trellis,
      "turbo",          turbo,
      NULL);

  if (bitrate > 0)
    ogmrip_encoding_set_bitrate (encoding, bitrate * 1000);

  ogmrip_encoding_set_scale (encoding, scale, scale_w, scale_h);
  ogmrip_encoding_set_crop (encoding, crop, crop_x, crop_y, crop_w, crop_h);

  if (fourcc)
    ogmrip_encoding_set_fourcc (encoding, fourcc);

  if (acodec == G_TYPE_NONE)
    acodec = ogmrip_plugin_get_nth_audio_codec (0);

  convert_alang (title);

  ogmrip_audio_options_init (&options.audio_options);
  options.audio_options.codec = acodec;
  options.audio_options.quality = audio_quality;
  options.audio_options.srate = 0 /* audio_srate */;
  options.audio_options.normalize = normalize;

  for (i = 0; i < 100; i++)
  {
    if (aid[i])
    {
      astream = ogmdvd_title_get_nth_audio_stream (title, i);

      if (!ogmrip_encoding_add_audio_stream (encoding, astream, &options.audio_options, error))
      {
        g_object_unref (encoding);
        return NULL;
      }
    }
  }

  ogmrip_audio_options_reset (&options.audio_options);

  if (scodec == G_TYPE_NONE)
    scodec = ogmrip_plugin_get_nth_subp_codec (0);

  convert_slang (title);

  ogmrip_subp_options_init (&options.subp_options);
  options.subp_options.codec = scodec;
  options.subp_options.charset = charset;
  options.subp_options.newline = newline;
  options.subp_options.forced_subs = forced_sub;

  for (i = 0; i < 100; i++)
  {
    if (sid[i])
    {
      sstream = ogmdvd_title_get_nth_subp_stream (title, i);

      if (!ogmrip_encoding_add_subp_stream (encoding, sstream, &options.subp_options, error))
      {
        g_object_unref (encoding);
        return NULL;
      }
    }
  }

  g_signal_connect (encoding, "task", G_CALLBACK (encoding_callback), NULL);

  return encoding;
}

