/* 
 * shrip - A DVD Encoder 
 * Copyright (C) 2004-2009 Olivier Rolland <billl@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "shrip-help.h"
#include "shrip-common.h"

#include <stdlib.h>

#include <glib.h>
#include <glib/gi18n.h>
#include <glib/gstdio.h>

#include <ogmdvd.h>
#include <ogmjob.h>
#include <ogmrip.h>

extern gchar    *device;
static gint      ntitle    = -1;
static gboolean  video     = FALSE;
static gboolean  audio     = FALSE;
static gboolean  chapters  = FALSE;
static gboolean  angles    = FALSE;
static gboolean  palette   = FALSE;
static gboolean  subp      = FALSE;
static gboolean  all       = FALSE;
static gboolean  help      = FALSE;
static gboolean  debug     = FALSE;

#ifdef OGMRIP_CHECK_VERSION
#if OGMRIP_CHECK_VERSION(0,11,0)
#define ogmdvd_audio_stream_unref(astream) ogmdvd_stream_unref (OGMDVD_STREAM (astream))
#define ogmdvd_subp_stream_unref(sstream)  ogmdvd_stream_unref (OGMDVD_STREAM (sstream))
#endif
#endif

static void
video_info (OGMDvdTitle *title)
{
  guint width, height;
  gint val;

  val = ogmdvd_title_get_video_format (title);
  g_print ("\t%s: %s, ", _("Format"), ogmdvd_get_video_format_label (val));

  val = ogmdvd_title_get_display_aspect (title);
  g_print ("%s: %s, ", _("Aspect ratio"), ogmdvd_get_display_aspect_label (val));

  ogmdvd_title_get_size (title, &width, &height);
  g_print ("%s: %dx%d\n", _("Size"), width, height);
}

static void
palette_info (OGMDvdTitle *title)
{
  const guint *palette;
  guint i;

  palette = ogmdvd_title_get_palette (title);
  g_print ("\t%s:", ("Palette"));
  for (i = 0; i < 16; i++)
    printf (" %06x", palette[i]);
  printf ("\n");
}

static void
audio_info (OGMDvdTitle *title)
{
  OGMDvdAudioStream *astream;
  gint aid, naid, val;

  naid = ogmdvd_title_get_n_audio_streams (title);
  for (aid = 0; aid < naid; aid ++)
  {
    astream = ogmdvd_title_get_nth_audio_stream (title, aid);
    if (astream)
    {
      g_print ("\t%s: %02d, ", _("Audio"),aid + 1);

      val = ogmdvd_audio_stream_get_language (astream);
      g_print ("%s: %s, ", _("Language"), ogmdvd_get_language_iso639_1 (val));

      val = ogmdvd_audio_stream_get_format (astream);
      g_print ("%s: %s, ", _("Format"), ogmdvd_get_audio_format_label (val));

      g_print ("%s: 48 kHz, ", _("Frequency"));

      val = ogmdvd_audio_stream_get_quantization (astream);
      g_print ("%s: %s, ", _("Quantization"), ogmdvd_get_audio_quantization_label (val));

      val = ogmdvd_audio_stream_get_channels (astream);
      g_print ("%s: %s, ", _("Channels"), ogmdvd_get_audio_channels_label (val));

      val = ogmdvd_audio_stream_get_content (astream);
      g_print ("%s: %s\n", _("Content"), ogmdvd_get_audio_content_label (val));

      ogmdvd_audio_stream_unref (astream);
    }
  }
}

static void
subp_info (OGMDvdTitle *title)
{
  OGMDvdSubpStream *sstream;
  gint sid, nsid, val;

  nsid = ogmdvd_title_get_n_subp_streams (title);
  for (sid = 0; sid < nsid; sid ++)
  {
    sstream = ogmdvd_title_get_nth_subp_stream (title, sid);
    if (sstream)
    {
      g_print ("\t%s: %02d, ", _("Subtitle"), sid + 1);

      val = ogmdvd_subp_stream_get_language (sstream);
      g_print ("%s: %s, ", _("Language"), ogmdvd_get_language_iso639_1 (val));

      val = ogmdvd_subp_stream_get_content (sstream);
      g_print ("%s: %s\n", _("Content"), ogmdvd_get_subp_content_label (val));

      ogmdvd_subp_stream_unref (sstream);
    }
  }
}

static void
chapters_info (OGMDvdTitle *title)
{
  OGMDvdTime time_;
  gint chap, nchap;

  nchap = ogmdvd_title_get_n_chapters (title);
  for (chap = 0; chap < nchap; chap ++)
  {
    ogmdvd_title_get_chapters_length (title, chap, chap, &time_);
    g_print ("\t\t%s: %02d, %s: %02d:%02d:%02d\n", 
        _("Chapter"), chap + 1, _("Length"), time_.hour, time_.min, time_.sec);
  }
}

static gint
title_info (OGMDvdTitle *title)
{
  OGMDvdTime time_;

  gint num_chapters, num_audio, num_subp;

  ogmdvd_title_get_length (title, &time_);
  num_chapters = ogmdvd_title_get_n_chapters (title);
  num_audio = ogmdvd_title_get_n_audio_streams (title);
  num_subp = ogmdvd_title_get_n_subp_streams (title);

  g_print ("%s: %02d, ", _("Title"), ntitle);
  g_print ("%s: %02d:%02d:%02d, ", _("Length"), time_.hour, time_.min, time_.sec);
  g_print ("%s: %02d, ", _("Chapters"), num_chapters);
  g_print ("%s: %02d, ", _("Audio streams"), num_audio);
  g_print ("%s: %02d\n", _("Subpictures"), num_subp);

  if (video)
    video_info (title);
  
  if (palette)
    palette_info (title);

  if (audio)
    audio_info (title);

  if (subp)
    subp_info (title);

  if (angles)
    g_print ("\t%s: %d\n", _("Number of angles"), ogmdvd_title_get_n_angles (title));

  if (chapters)
    chapters_info (title);

  return EXIT_SUCCESS;
}

static gint
dvd_info (OGMDvdDisc *disc)
{
  OGMDvdTitle *title;
  gdouble length, longest_length;
  gint longest_title, ntitles;

  ntitles = ogmdvd_disc_get_n_titles (disc);

  for (longest_title = ntitle = 1, longest_length = 0; ntitle <= ntitles; ntitle++)
  {
    title = ogmdvd_disc_get_nth_title (disc, ntitle - 1);
    if (!title)
    {
      g_printerr (_("Cannot open title %d."), ntitle);
      g_printerr ("\n");
      return EXIT_FAILURE;
    }

    title_info (title);

    length = ogmdvd_title_get_length (title, NULL);
    if (length > longest_length)
    {
      longest_length = length;
      longest_title = ntitle;
    }

    ogmdvd_title_unref (title);
  }
  g_print ("%s: %d\n", _("Longest title"), longest_title);

  return EXIT_SUCCESS;
}

static const GOptionEntry entries[] = 
{
  { "device",     0,  0, G_OPTION_ARG_CALLBACK, parse_device, N_("Specify the DVD device (default: /dev/dvd)"), "device" },
  { "title",      0,  0, G_OPTION_ARG_INT,      &ntitle,   N_("Specify the DVD title (default: none)"),      "id"     },
  { "video",      0,  0, G_OPTION_ARG_NONE,     &video,    N_("Show video info"),                            NULL     },
  { "audio",      0,  0, G_OPTION_ARG_NONE,     &audio,    N_("Show audio info"),                            NULL     },
  { "chapters",   0,  0, G_OPTION_ARG_NONE,     &chapters, N_("Show chapters info"),                         NULL     },
  { "angles",     0,  0, G_OPTION_ARG_NONE,     &angles,   N_("Show angles info"),                           NULL     },
  { "palette",    0,  0, G_OPTION_ARG_NONE,     &palette,  N_("Show palette info"),                          NULL     },
  { "subtitles",  0,  0, G_OPTION_ARG_NONE,     &subp,     N_("Show subtitles info"),                        NULL     },
  { "all",        0,  0, G_OPTION_ARG_NONE,     &all,      N_("Show all info"),                              NULL     },
  { NULL,         0,  0, 0,                     NULL,      NULL,                                             NULL     }
};

static const GOptionEntry help_entries[] =
{
  { "help", 'h', 0, G_OPTION_ARG_NONE, &help,  N_("Show help options"),     NULL },
  { "debug", 0,  0, G_OPTION_ARG_NONE, &debug, N_("Enable debug messages"), NULL },
  { NULL,    0,  0, 0,                 NULL,  NULL,                         NULL }
};

static void
load_filename_default (GKeyFile *key_file, const gchar *group_name, const gchar *key_name, gchar **value)
{
  GError *error = NULL;
  gchar *val;

  val = g_key_file_get_locale_string (key_file, group_name, key_name, "UTF8", &error);
  if (error)
    g_error_free (error);
  else
  {
    *value = g_filename_from_utf8 (val, -1, NULL, NULL, NULL);
    g_free (val);
  }
}

static void
load_defaults (const gchar *filename)
{
  GKeyFile *key_file;

  key_file = g_key_file_new ();
  if (g_key_file_load_from_file (key_file, filename, G_KEY_FILE_NONE, NULL))
    load_filename_default (key_file, "main", "device", &device);
  g_key_file_free (key_file);
}

static void
init (gint *argc, gchar ***argv)
{
  (*argv)[0] = "shrip list";
}

gint
list (gint *argc, gchar ***argv)
{
  OGMDvdDisc *disc = NULL;
  GError *error = NULL;
  gchar *filename;

  GOptionContext *context;
  GOptionGroup   *group;

  gint retval = EXIT_FAILURE;

  init (argc, argv);

  filename = g_build_filename (SHRIP_CONF_DIR, "shrip.conf", NULL);
  load_defaults (filename);
  g_free (filename);

  filename = g_build_filename (g_get_home_dir (), ".shrip", NULL);
  load_defaults (filename);
  g_free (filename);

  context = g_option_context_new (NULL);
  g_option_context_set_help_enabled (context, FALSE);
  g_option_context_add_main_entries (context, entries, NULL);

  group = g_option_group_new ("help", _("Help Options"), _("Show help options"), NULL, NULL);
  g_option_group_add_entries (group, help_entries);
  g_option_context_add_group (context, group);

  if (!g_option_context_parse (context, argc, argv, &error))
  {
    g_option_context_free (context);
    goto cleanup;
  }
  g_option_context_free (context);

  if (help)
  {
    print_help ("list", NULL, _("Help Options"), help_entries, _("List Options"), entries, NULL);
    return EXIT_SUCCESS;
  }

#ifdef OGMRIP_CHECK_VERSION
#if OGMRIP_CHECK_VERSION(0,11,0)
  if (debug)
    ogmjob_log_set_print_stdout (TRUE);
#endif
#endif

  if (all)
    video = audio = chapters = angles = palette = subp = TRUE;

  if (!device)
    device = "/dev/dvd";

  disc = ogmdvd_disc_new (device, NULL);
  if (!disc || !ogmdvd_disc_open (disc, NULL))
  {
    g_printerr (_("Couldn't open DVD device: '%s'"), device);
    g_printerr ("\n");
    goto cleanup;
  }

  if (ntitle == -1)
    retval = dvd_info (disc);
  else
  {
    OGMDvdTitle *title;

    if (ntitle < 1 || ntitle > ogmdvd_disc_get_n_titles (disc))
    {
      g_printerr (_("The title %d does not exist."), ntitle);
      g_printerr ("\n");
      goto cleanup;
    }

    title = ogmdvd_disc_get_nth_title (disc, ntitle - 1);
    if (!title)
    {
      g_printerr (_("Cannot open title %d."), ntitle);
      g_printerr ("\n");
      goto cleanup;
    }

    retval = title_info (title);

    ogmdvd_title_unref (title);
  }

cleanup:
  if (disc)
    ogmdvd_disc_unref (disc);

  if (error)
  {
    g_printerr ("%s\n", error->message);
    g_error_free (error);
  }

  return retval;
}

